/*----------------------------------------------------------------------
  File    : lvqd.c
  Contents: learning vector quantization demonstration
  Author  : Christian Borgelt
  History : 2000.09.01 file created from file xlvq.c
            2000.09.02 function lvq_load added
            2001.06.27 bug in function lvq_load removed
            2001.11.08 table functions moved to module tab4vis
            2003.09.02 (fuzzy) c-means clustering added
            2013.08.08 adapted to definitions ATTID, VALID, WEIGHT etc.
            2013.08.09 adapted to higher compiler warning level
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <assert.h>
#include "lvqd.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE       32        /* block size for history vector */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- add. reference vector info.--- */
  double cnt;                   /* number of assigned patterns */
  double msd;                   /* membership degree */
} RVADD;                        /* (add. reference vector info.) */

/*----------------------------------------------------------------------
  Global Variables
----------------------------------------------------------------------*/
/* --- neural network variables --- */
REFVEC *rvecs  = NULL;         /* current reference vectors and */
VALID  *rvcls  = NULL;         /* classes they are assigned to */
VALID  rvcnt   = 0;            /* number of reference vectors */
REFVEC **hist  = NULL;         /* history of reference vectors */
int    histvsz = 0;            /* size of history vector */
int    histcnt = 0;            /* num. of entries in history vector */
RVADD  *rvadd  = NULL;         /* add. reference vector info. */
LVQPAR lvqpar  =               /* l.v.q. parameters */
  { LVQ_CENTER, 1, 0.1, 0.001, -0.1, 0.01 };
static int init = 0;            /* initialization flag */

/* --- variables for loading ref. vectors --- */
static struct {                 /* --- new reference vectors --- */
  ATTID  xid, yid, clsid;       /* attribute identifiers */
  VALID  cnt;                   /* number of reference vectors */
  VALID  size;                  /* current array size */
  REFVEC *rvecs;                /* new reference vectors and */
  VALID  *rvcls;                /* the classes assigned to them */
} nvs;                          /* (new reference vectors) */

/*----------------------------------------------------------------------
  Functions
----------------------------------------------------------------------*/

static double drand (void)
{ return (double)rand()/((double)RAND_MAX+1.0); }

/*--------------------------------------------------------------------*/

void lvq_clean (void)
{                               /* --- clean up the reference vectors */
  int i;                        /* loop variable */

  if (hist) {                   /* if there is a vector history */
    for (i = 0; i < histcnt; i++) free(hist[i]);
    free(hist); hist = NULL;    /* delete all reference vectors, */
    histcnt = histvsz = 0;      /* the history vector, and */
  }                             /* the current reference vectors */
  if (rvadd) { free(rvadd); rvadd = NULL; }
  if (rvcls) { free(rvcls); rvcls = NULL; }
  if (rvecs) { free(rvecs); rvecs = NULL; }
}  /* lvq_clean() */

/*--------------------------------------------------------------------*/

int lvq_save (const char *fname)
{                               /* --- save reference vectors */
  VALID i;                      /* loop variable */
  ATT   *att;                   /* class attribute */
  FILE  *out;                   /* output file */

  assert(fname);                /* check the function argument */
  assert(attset && rvecs);      /* check l.v.q. variables */
  att = (selatt.c_att < 0)      /* get the class attribute */
      ? NULL : as_att(attset, map_nom[selatt.c_att]);
  out = fopen(fname, "w");      /* open the output file */
  if (!out) return E_FOPEN;     /* and check for success */
  fputs(nms_met[selatt.h_att], out); fputc('\n', out);
  fputs(nms_met[selatt.v_att], out); fputc('\n', out);
  if (att) fputs(nms_nom[selatt.c_att], out);
  fputc('\n', out);             /* print the attribute names */
  for (i = 0; i < rvcnt; i++) { /* traverse the reference vectors */
    fprintf(out, "%g %g", rvecs[i].x, rvecs[i].y);
    if (att) { fputc(' ', out); fputs(att_valname(att,rvcls[i]), out); }
    fputc('\n', out);           /* print the vector coordinates */
  }                             /* and the class name */
  return (fclose(out) != 0) ? E_FWRITE : 0;
}  /* lvq_save() */             /* close the output file */

/*--------------------------------------------------------------------*/

static int load (TABREAD *tread)
{                               /* --- load reference vectors */
  int    type;                  /* attribute type */
  VALID  cls = 0;               /* class of the reference vector */
  double x, y;                  /* coordinates of a vector */
  ATT    *att = NULL;           /* class attribute (if any) */
  int    d;                     /* delimiter type */
  VALID  n;                     /* size of new vector */
  char   *s, *e;                /* read field and end pointer */
  void   *p;                    /* buffer for reallocation */

  nvs.rvecs = NULL;             /* clear the variables, in which */
  nvs.rvcls = NULL;             /* the loaded reference vectors */
  nvs.size  = nvs.cnt = 0;      /* will be buffered, and */
  nvs.clsid = -1;               /* clear the class identifier */

  d = trd_read(tread);          /* get name of the horz. attribute */
  if (d != TRD_REC) return E_FREAD;
  nvs.xid = as_attid(attset, trd_field(tread));
  if (nvs.xid < 0)  return E_RVINC; /* get attribute identifier */
  type = att_type(as_att(attset, nvs.xid));
  if ((type != AT_FLT) && (type != AT_INT))
    return E_RVINC;             /* check the attribute type (metric) */

  d = trd_read(tread);          /* get name of the vert. attribute */
  if (d != TRD_REC) return E_FREAD;
  nvs.yid = as_attid(attset, trd_field(tread));
  if (nvs.yid < 0)  return E_RVINC; /* get attribute identifier */
  type = att_type(as_att(attset, nvs.yid));
  if ((type != AT_FLT) && (type != AT_INT))
    return E_RVINC;             /* check the attribute type (metric) */

  d = trd_read(tread);          /* read the next field */
  if (d != TRD_REC) return E_FREAD;
  s = trd_field(tread);         /* get name of the class attribute */
  if (*s != '\0') {             /* if a class attribute is specified */
    nvs.clsid = as_attid(attset, s);   /* look up the att. name */
    if (nvs.clsid < 0) return E_RVINC; /* in the attribute set and */
    att = as_att(attset, nvs.clsid);   /* get the attribute id. */
    if (att_type(att) != AT_NOM) return E_RVINC;
  }                             /* the class att. must be nominal */

  while (1) {                   /* reference vector read loop */
    d = trd_read(tread);        /* read the x-coordinate */
    if (d == TRD_EOF) break;    /* check for end of file */
    if (d != TRD_FLD)   return E_FREAD;
    x = strtod(s = trd_field(tread), &e);
    if ((e == s) || *e) return E_FREAD;
    d = trd_read(tread);        /* read the y-coordinate */
    if (d <= TRD_EOF)   return E_FREAD;
    y = strtod(s = trd_field(tread), &e);
    if ((e == s) || *e) return E_FREAD;
    if (!att) {                 /* if no class att. is specified, */
      if (d != TRD_REC) return E_FREAD; }  /* check for rec. sep. */
    else {                      /* if a class attribute is specified */
      d = trd_read(tread);      /* read the class value */
      if (d != TRD_REC) return E_FREAD;
      cls = att_valid(att, trd_field(tread));
      if (cls < 0)      return E_RVINC;
    }                           /* get the class identifier */
    n = nvs.size;               /* get the r.v. array size */
    if (nvs.cnt >= n) {         /* if the r.v. array is full, */
      n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
      p = realloc(nvs.rvecs, (size_t)n *sizeof(REFVEC));
      if (!p) return E_NOMEM;   /* enlarge the r.v. vector */
      nvs.rvecs = (REFVEC*)p;   /* and replace the old one */
      p = realloc(nvs.rvcls, (size_t)n *sizeof(VALID));
      if (!p) return E_NOMEM;   /* enlarge the class vector */
      nvs.rvcls = (VALID*)p;    /* and replace the old one */
      nvs.size  = n;            /* set the new varray size */
    }
    nvs.rvecs[nvs.cnt].x = x;   /* store the loaded */
    nvs.rvecs[nvs.cnt].y = y;   /* reference vector */
    nvs.rvcls[nvs.cnt] = cls;   /* and its class */
    nvs.cnt++;                  /* count the reference vector */
  }
  if (nvs.cnt <= 0) return E_FREAD;

  return 0;                     /* return 'ok' */
}  /* load() */

/*--------------------------------------------------------------------*/

int lvq_load (const char *fname)
{                               /* --- load reference vectors */
  TABREAD *tread;               /* table file reader */
  int     r;                    /* result of load function */

  assert(fname);                /* check the function argument */
  if (!attset) return 0;        /* check for an attribute set */
  tread = trd_create();         /* create a table file reader */
  if (!tread) return E_NOMEM;   /* and open the input file */
  if (trd_open(tread, NULL, fname) != 0) {
    trd_delete(tread, 1); return E_FOPEN; }
  r = load(tread);              /* load reference vectors */
  if (r != 0) {                 /* clean up on error */
    if (nvs.rvecs) free(nvs.rvecs);
    if (nvs.rvcls) free(nvs.rvcls); }
  else {                        /* if loading was successful, */
    lvq_clean();                /* delete the old vectors */
    rvecs = nvs.rvecs;          /* set the new reference vectors, */
    rvcls = nvs.rvcls;          /* their assigned classes, */
    rvcnt = nvs.cnt;            /* and their number */
    selatt.h_att = inv_met[nvs.xid];
    selatt.h_rng = ranges[selatt.h_att];
    selatt.v_att = inv_met[nvs.yid];
    selatt.v_rng = ranges[selatt.v_att];
    selatt.c_att = (nvs.clsid >= 0) ? inv_nom[nvs.clsid] : -1;
  }                             /* note the selected attributes */
  trd_delete(tread, 1);         /* delete the table file reader */
  return r;                     /* return 'ok' */
}  /* lvq_load() */

/*--------------------------------------------------------------------*/

int lvq_init (void)
{                               /* --- initialize reference vectors */
  int    i, k, n = 0;           /* loop variables, indices */
  float  f;                     /* temporary buffer */
  int    clscnt, cls;           /* number of classes, current class */
  int    xid, yid, clsid;       /* attribute identifiers */
  int    xtype, ytype;          /* attribute types */
  double x, y;                  /* coordinates of data point */
  int    *cnts;                 /* counters for different purposes */
  RANGE  *rng;                  /* to traverse the attribute ranges */
  TUPLE  *tpl;                  /* to traverse the data tuples */

  lvq_clean();                  /* delete old vectors */
  if (!table) return 0;         /* check for a data table */
  if (!init) {                  /* init. random number generator */
    srand((unsigned int)time(NULL)); init = 1; }

  /* --- create vectors --- */
  clscnt = (selatt.c_att < 0)   /* get the number of classes */
         ? 1 : att_valcnt(as_att(attset, map_nom[selatt.c_att]));
  rvcnt = clscnt *lvqpar.vpcls; /* compute the number of vectors */
  rvecs = (REFVEC*)calloc((size_t)rvcnt,  sizeof(REFVEC));
  rvcls = (VALID*) malloc((size_t)rvcnt  *sizeof(VALID));
  cnts  = (VALID*) calloc((size_t)clscnt, sizeof(VALID));
  if (!rvecs || !rvcls || !cnts) {
    if (cnts) free(cnts);       /* create buffers for vectors, */
    lvq_clean(); return E_NOMEM;/* classes etc. */
  }

  /* --- initialize vectors --- */
  if (lvqpar.imode == LVQ_CENTER) {  /* if center initialization */
    rng = ranges +selatt.h_att; /* get horizontal center value */
    x   = 0.5 *(rng->max +rng->min);
    rng = ranges +selatt.v_att; /* get vertical   center value */
    y   = 0.5 *(rng->max +rng->min);
    for (k = 0; k < clscnt; k++) {
      for (i = lvqpar.vpcls; --i >= 0; ) {
        rvecs[n].x = x;         /* traverse the classes and */
        rvecs[n].y = y;         /* for each class set the requested */
        rvcls[n++] = k;         /* number of vectors to the center */
      }                         /* of the range of values */
    } }
  else {                        /* if data driven initialization */
    if (lvqpar.imode == LVQ_DATAPT) { /* if random data point init. */
      tab_shuffle(table, 0, TPLID_MAX, drand);
      for (i = clscnt; --i >= 0; )
        cnts[i] = lvqpar.vpcls; /* shuffle the data tuples and */
    }                           /* initialize the vector counters */
    clsid = (selatt.c_att < 0) ? -1 : map_nom[selatt.c_att];
    xid   = map_met[selatt.h_att];  /* get attribute identifiers */
    yid   = map_met[selatt.v_att];  /* and attribute types */
    xtype = att_type(as_att(attset, xid));
    ytype = att_type(as_att(attset, yid));
    cls   = 0;                  /* get the default class (none) */
    for (k = tab_tplcnt(table); --k >= 0; ) {
      tpl = tab_tpl(table, k);  /* traverse the data tuples */
      if (clsid >= 0) {         /* if a class attribute is specified */
        cls = tpl_colval(tpl, clsid)->i;
        if (cls < 0) continue;  /* get the class of the tuple and */
      }                         /* skip tuples with an unknown class */
      if (xtype == AT_INT) {    /* if the x-coord. is integer-valued */
        i = tpl_colval(tpl, xid)->i;
        if (isnull(i)) continue;
        x = (double)i; }        /* get and check the attribute value */
      else {                    /* if the x-coord. is real-valued */
        f = tpl_colval(tpl, xid)->f;
        if (isnan(f))  continue;
        x = (double)f;          /* get and check the attribute value */
      }
      if (ytype == AT_INT) {    /* if the y-coord. is integer-valued */
        i = tpl_colval(tpl, yid)->i;
        if (isnull(i)) continue;
        y = (double)i; }        /* get and check the attribute value */
      else {                    /* if the y-coord. is real-valued */
        f = tpl_colval(tpl, yid)->f;
        if (isnan(f))  continue;
        y = (double)f;          /* get and check the attribute value */
      }
      if (lvqpar.imode == LVQ_DATAPT) {
        if (cnts[cls] <= 0)     /* if random data point init., */
          continue;             /* skip tuples with an unknown class */
        cnts[cls]--;            /* check the number of vectors still */
        rvecs[n].x = x;         /* needed for the class of the tuple */
        rvecs[n].y = y;         /* and if more vectors are needed, */
        rvcls[n++] = cls; }     /* copy the coordinates of the tuple */
      else {                    /* if class mean initialization */
        rvecs[cls].x += x;      /* sum the coordinates of the */
        rvecs[cls].y += y;      /* data tuples and count them */
        cnts[cls]++;            /* (for later class mean computation) */
      }
    }  /* for (k = tab_tplcnt(table); .. */

    if (lvqpar.imode == LVQ_MEAN) {
      x = y = 0; k = 0;         /* if global mean initialization */
      for (i = clscnt; --i >= 0; ) {
        x += rvecs[i].x;        /* traverse the classes */
        y += rvecs[i].y;        /* sum the class specific sums */
        k += cnts[i];           /* of the coordinates and */
      }                         /* the numbers of tuples */
      x /= k; y /= k;           /* compute the global means */
      for (i = clscnt; --i >= 0; ) {
        for (k = lvqpar.vpcls; --k >= 0; ) {
          rvecs[n].x = x;       /* traverse the reference vectors */
          rvecs[n].y = y;       /* and set each of them to the */
          rvcls[n++] = i;       /* computed global mean values */
        }                       /* and assign classes */
      } }
    else if (lvqpar.imode == LVQ_CLSMEAN) {
      for (i = clscnt; --i >= 0; ) { /* if class mean initialization */
        rvecs[i].x /= cnts[i];  /* traverse the summed coordinate */
        rvecs[i].y /= cnts[i];  /* values and compute the mean */
      }                         /* (divide by the number of tuples) */
      for (k = lvqpar.vpcls; --k >= 0; ) {
        for (i = 0; i < clscnt; i++) {
          rvecs[n].x = rvecs[i].x;
          rvecs[n].y = rvecs[i].y;
          rvcls[n++] = i;       /* traverse the reference vectors, */
        }                       /* set each of them to the mean */
      }                         /* value of a class and assign */
    }                           /* the corresponding class */
  }  /* if (lvqpar.imode == LVQ_CENTER) .. else .. */
  free(cnts);                   /* delete the counter vector */

  /* --- randomize reference vectors --- */
  if (lvqpar.roff > 0) {        /* if random offsets are to be added */
    rng = ranges +selatt.h_att;
    x   = rng->max -rng->min;   /* get horizontal range of values */
    rng = ranges +selatt.v_att;
    y   = rng->max -rng->min;   /* get vertical   range of values */
    for (i = rvcnt; --i >= 0; ) {
      rvecs[i].x += x *2 *lvqpar.roff *(drand() -0.5);
      rvecs[i].y += y *2 *lvqpar.roff *(drand() -0.5);
    }                           /* add random numbers to the vectors */
  }
  return 0;                     /* return 'ok' */
}  /* lvq_init() */

/*--------------------------------------------------------------------*/

int lvq_train (void)
{                               /* --- do one training step */
  int    i, k, best;            /* loop variables, index */
  float  f;                     /* temporary buffer */
  TUPLE  *tpl;                  /* to traverse the tuples */
  REFVEC *rvnew;                /* new set of reference vectors */
  REFVEC **p;                   /* buffer for memory (re)allocation */
  int    xid, yid, clsid;       /* attribute identifiers */
  int    xtype, ytype;          /* attribute types */
  int    cls;                   /* class of current tuple */
  double rx, ry;                /* coordinates of data point */
  double dx, dy, d;             /* distances to a reference vector */
  double cx, cy, min;           /* ditto, for closest ref. vector */
  double sum;                   /* sum of membership degrees */
  double lrate;                 /* current learning rate */

  if (!rvecs) return 0;         /* if there are no vectors, abort */
  if (!rvadd) {                 /* if there is no add. information */
    rvadd = (RVADD*)malloc((size_t)rvcnt *sizeof(RVADD));
    if (!rvadd) return -1;      /* create a vector of additional */
  }                             /* reference vector information */
  if (histcnt >= histvsz) {     /* if the history vector is full */
    k = histvsz +((histvsz > BLKSIZE) ? histvsz >> 1 : BLKSIZE);
    p = (REFVEC**)realloc(hist, (size_t)k *sizeof(REFVEC*));
    if (!p) return -1;          /* enlarge the history vector and */
    hist = p; histvsz = k;      /* set the new vector and its size */
  }
  rvnew = (REFVEC*)calloc((size_t)rvcnt, sizeof(REFVEC));
  if (!rvnew) return -1;        /* create new reference vectors */
  hist[histcnt++] = rvecs;      /* and add old ones to the history */
  for (i = rvcnt; --i >= 0; )   /* clear the number */
    rvadd[i].cnt = 0;           /* of assigned patterns */

  clsid = (selatt.c_att < 0) ? -1 : map_nom[selatt.c_att];
  xid   = map_met[selatt.h_att];/* get attribute identifiers */
  yid   = map_met[selatt.v_att];/* and attribute types */
  xtype = att_type(as_att(attset, xid));
  ytype = att_type(as_att(attset, yid));
  lrate =  (lvqpar.lrate <= 0) ? lvqpar.lrate
        : ((lvqpar.decay <= 0)
        ? fabs(lvqpar.lrate *pow(histcnt +1, lvqpar.decay))
        : fabs(lvqpar.lrate *pow(lvqpar.decay, histcnt +1)));
  cls = 0;                      /* compute the current learning rate */
  for (k = tab_tplcnt(table); --k >= 0; ) {
    tpl = tab_tpl(table, k);    /* traverse the data tuples */
    if (clsid >= 0) {           /* if a class attribute is specified */
      cls = tpl_colval(tpl, clsid)->i;
      if (cls < 0) continue;    /* get the class of the tuple and */
    }                           /* skip tuples with an unknown class */
    if (xtype == AT_INT) {      /* if x-coordinate is integer valued */
      i = tpl_colval(tpl, xid)->i;
      if (isnull(i)) continue;
      rx = (double)i; }         /* get and check the attribute value */
    else {                      /* if x-coordinate is real valued */
      f = tpl_colval(tpl, xid)->f;
      if (isnan(f))  continue;
      rx = (double)f;           /* get and check the attribute value */
    }
    if (ytype == AT_INT) {      /* if y-coordinate is integer valued */
      i = tpl_colval(tpl, yid)->i;
      if (isnull(i)) continue;
      ry = (double)i; }         /* get and check the attribute value */
    else {                      /* if y-coordinate is real valued */
      f = tpl_colval(tpl, yid)->f;
      if (isnan(f))  continue;
      ry = (double)f;           /* get and check the attribute value */
    }
    best = -1; min = INFINITY;  /* initialize the minimal distance */
    cx = cy = sum = 0;          /* and traverse the reference vectors */
    for (i = rvcnt; --i >= 0; ) {
      dx = rx -rvecs[i].x;      /* compute the Euclidean distance */
      dy = ry -rvecs[i].y;      /* between the current data point and */
      d  = dx*dx +dy*dy;        /* the different reference vectors */
      if (lrate < -1)           /* compute the membership degree */
        sum += rvadd[i].msd = (d > 1e-12) ? pow(d, lrate/2) : 1e12;
      if (d < min) { min = d; cx = dx; cy = dy; best = i; }
    }                           /* determine the closest ref. vector */
    if (best < 0) continue;     /* check for a successful search */
    if (lrate < -1) {           /* if fuzzy c-means clustering */
      sum = (sum > 0) ? 1/sum : 1; /* compute the norm. factor */
      for (i = rvcnt; --i >= 0; ) {
        rvadd[i].cnt += d = rvadd[i].msd *sum;
        rvnew[i].x += (rx -rvecs[i].x) *d;
        rvnew[i].y += (ry -rvecs[i].y) *d;
      } }                       /* aggregate the center changes */
    else {                      /* if LVQ or c-means clustering */
      if (lrate <= 0) d = 1;    /* det. the sign of the learning rate */
      else            d = (cls == rvcls[best]) ? lrate : -lrate;
      rvnew[best].x += cx *d;   /* aggregate the changes that have to */
      rvnew[best].y += cy *d;   /* be made to the reference vectors */
      rvadd[best].cnt += 1;     /* and count the training pattern */
    }
  }
  for (i = rvcnt; --i >= 0; ) { /* traverse the reference vectors and */
    if (lrate <= 0) {           /* if (fuzzy) c-means clustering */
      rvnew[i].x /= rvadd[i].cnt;
      rvnew[i].y /= rvadd[i].cnt;
    }                           /* compute change to new center */
    rvnew[i].x += rvecs[i].x;   /* add the current reference vectors */
    rvnew[i].y += rvecs[i].y;   /* to the aggregated changes to */
  }                             /* obtain the new reference vectors */
  rvecs = rvnew;                /* set the new reference vectors */
  return 0;                     /* return 'ok' */
}  /* lvq_train() */
